/*-----------------------------------------------------------------------*/
/* Low level disk I/O module SKELETON for FatFs     (C)ChaN, 2019        */
/*-----------------------------------------------------------------------*/
/* If a working storage control module is available, it should be        */
/* attached to the FatFs via a glue function rather than modifying it.   */
/* This is an example of glue functions to attach various exsisting      */
/* storage control modules to the FatFs module with a defined API.       */
/*-----------------------------------------------------------------------*/

// Modified to work with MCU SDIO peripherals

#include "ff.h"			/* Obtains integer types */
#include "diskio.h"		/* Declarations of disk functions */
#include "sd_card_protocol.h"



volatile DSTATUS DiskStatus = STA_NOINIT | STA_NODISK;

/*-----------------------------------------------------------------------*/
/* Get Drive Status                                                      */
/*-----------------------------------------------------------------------*/

DSTATUS disk_status (
	BYTE pdrv		/* Physical drive number to identify the drive */
)
{
	if (pdrv)
		return STA_NOINIT;		// Supports only drive 0

	return DiskStatus;	    // Return disk status
}



/*-----------------------------------------------------------------------*/
/* Inidialize a Drive                                                    */
/*-----------------------------------------------------------------------*/

DSTATUS disk_initialize (
	BYTE pdrv				/* Physical drive number to identify the drive */
)
{
	SD_Detect_Result_TypeDef res;

	if (pdrv) return STA_NOINIT;

	res = SD_CARD_CardDetect();

	if (res == SD_OK)
		DiskStatus = 0;
	else if (SD_CARD_IsCardInsert())
		DiskStatus = STA_NOINIT;
	else
		DiskStatus = STA_NOINIT | STA_NODISK;

	return DiskStatus;
}



/*-----------------------------------------------------------------------*/
/* Read Sector(s)                                                        */
/*-----------------------------------------------------------------------*/

DRESULT disk_read (
	BYTE pdrv,		/* Physical drive number to identify the drive */
	BYTE *buff,		/* Data buffer to store read data */
	LBA_t sector,	/* Start sector in LBA */
	UINT count		/* Number of sectors to read */
)
{
	if (pdrv || !count) return RES_PARERR;			// Check parameter
	if (DiskStatus & STA_NOINIT) return RES_NOTRDY;	// Check if drive is ready

	return (DRESULT)SD_CARD_CardRead(buff, sector, count);
}



/*-----------------------------------------------------------------------*/
/* Write Sector(s)                                                       */
/*-----------------------------------------------------------------------*/

#if FF_FS_READONLY == 0

DRESULT disk_write (
	BYTE pdrv,			/* Physical drive number to identify the drive */
	const BYTE *buff,	/* Data to be written */
	LBA_t sector,		/* Start sector in LBA */
	UINT count			/* Number of sectors to write */
)
{
	if (pdrv || !count) return RES_PARERR;			/* Check parameter */
	if (DiskStatus & STA_NOINIT) return RES_NOTRDY;	/* Check drive status */
	if (DiskStatus & STA_PROTECT) return RES_WRPRT;	/* Check write protect */

	return (DRESULT)SD_CARD_CardWrite(buff, sector, count);
}

#endif


/*-----------------------------------------------------------------------*/
/* Miscellaneous Functions                                               */
/*-----------------------------------------------------------------------*/

DRESULT disk_ioctl (
	BYTE pdrv,		/* Physical drive number (0..) */
	BYTE cmd,		/* Control code */
	void *buff		/* Buffer to send/receive control data */
)
{
	volatile DRESULT res;

	if (pdrv) return RES_PARERR;					/* Check parameter */
	if (DiskStatus & STA_NOINIT) return RES_NOTRDY;	/* Check drive status */  

	//  SD-Card disk_ioctl
	res = (DRESULT) SD_CARD_IO_Ctrl((SD_Card_IO_Ctrl_TypeDef)cmd, buff);

	if (!SD_CARD_IsCardReady())
		DiskStatus |= STA_NOINIT;

	return res;
}

