#include "MDR32FxQI_bkp.h"
#include "MDR32FxQI_eeprom.h"
#include "MDR32FxQI_port.h"
#include "MDR32FxQI_rst_clk.h"
#include "MDR32FxQI_sdio.h"

#include "sd_card_sdio_ll.h"
#include "sd_card_protocol.h"
#include <stdio.h>

// FileSystem
#include "test_funcIO.h"

// Status indicating LEDs
#define LED_OK      PORT_Pin_15 // PB15 -  VD3
#define LED_FAULT   PORT_Pin_14 // PB14 -  VD4
#define LED_CYCLE   PORT_Pin_13 // PB13 -  VD5

#define LED_SEL_ALL PORT_Pin_15 | PORT_Pin_14 | PORT_Pin_13

// Clock settings
// SysTick with 1ms period is used for SD card timeouts handling
// SysTick timer starts counting in SDIO_LL_Init() which is called in SD_CARD_Init()
#define PLL_MULL   RST_CLK_CPU_PLLmul10
#define PERIOD_1MS (HSE_Value * (PLL_MULL + 1) / 1000)

uint32_t TmrDelay_ms = 1000;

// Tests in cycle - only one could be active
void While_TestFS(void);
void While_SDIO_TestCMD(void);
void While_SDIO_TestCMD_8Clk(void);
void While_SDIO_TestCycles(void);
void While_SDIO_TestRate(void);

// Timeouts for SDIO peripheral functions
SD_CARD_Timeouts_TypeDef sdioTimeouts = {
    .SD_SendCMD_Timout_ms    = 100,
    .SD_BeforeRead_Delay_ms  = 2,
    .SD_ReadResp_Timout_ms   = 100,
    .SD_ChangeRate_Timout_ms = 2,
    .SD_ReadBlock_Timout_ms  = 100,
    .SD_WriteBlock_Timout_ms = 1000
};

// Timeouts and settings for SD card work
SD_CARD_ClockCfg_TypeDef sdioCardCfg = {
    .SysTimerPeriod_1ms  = PERIOD_1MS,      // Value to get 1ms timer period
    .SD_CLK_BRG_Detect   = SDIO_BRG_div256, // 80MHz/256 = 312,5 KHz < 400KHz
    .SD_CLK_BRG_Transfer = SDIO_BRG_div16,  // 80MHz/16  = 5 MHz

    .SD_PowerOn_Delay_ms     = 2,    //     2 ms
    .SD_WaitStatus_Timout_ms = 500,  //   500 ms
    .SD_Erase_Timout_ms      = 30000 // 30000 ms
};

PORT_InitTypeDef GPIOInitStruct;

int main(void)
{
    //  Clock 80MHz
    RST_CLK_DeInit();
    RST_CLK_PCLKcmd(RST_CLK_PCLK_EEPROM, ENABLE);
    RST_CLK_HSEconfig(RST_CLK_HSE_ON);
    while (RST_CLK_HSEstatus() == ERROR) { }
    RST_CLK_CPU_PLLconfig(RST_CLK_CPU_PLLsrcHSEdiv1, RST_CLK_CPU_PLLmul10);
    RST_CLK_CPU_PLLcmd(ENABLE);
    while (RST_CLK_CPU_PLLstatus() == ERROR) { }
    RST_CLK_CPUclkPrescaler(RST_CLK_CPUclkDIV1);
    RST_CLK_CPU_PLLuse(ENABLE);
    BKP_DUccMode(BKP_DUcc_upto_80MHz);
    EEPROM_SetLatency(EEPROM_Latency_3);
    RST_CLK_CPUclkSelection(RST_CLK_CPUclkCPU_C3);

    // Board LEDs configuration

    RST_CLK_PCLKcmd(RST_CLK_PCLK_PORTB, ENABLE);

    PORT_StructInit(&GPIOInitStruct);
    //                              VD3           VD4           VD5            VD6          VD7
    GPIOInitStruct.PORT_Pin   = (PORT_Pin_15 | PORT_Pin_14 | PORT_Pin_13 | PORT_Pin_12 | PORT_Pin_11);
    GPIOInitStruct.PORT_OE    = PORT_OE_OUT;
    GPIOInitStruct.PORT_SPEED = PORT_SPEED_SLOW;
    GPIOInitStruct.PORT_MODE  = PORT_MODE_DIGITAL;
    PORT_Init(MDR_PORTB, &GPIOInitStruct);

    // Board buttons configuration

    RST_CLK_PCLKcmd(RST_CLK_PCLK_PORTC, ENABLE);

    PORT_StructInit(&GPIOInitStruct);
    //                               SEL          UP          DOWN         LEFT        RIGHT
    GPIOInitStruct.PORT_Pin   = (PORT_Pin_0 | PORT_Pin_3 | PORT_Pin_4 | PORT_Pin_5 | PORT_Pin_6);
    GPIOInitStruct.PORT_SPEED = PORT_SPEED_SLOW;
    GPIOInitStruct.PORT_MODE  = PORT_MODE_DIGITAL;
    PORT_Init(MDR_PORTC, &GPIOInitStruct);

    // SD card initialiation
    SD_CARD_Init(&sdioCardCfg, &sdioTimeouts);

    // Setup SysTick for timeouts handling
    SysTick->LOAD = PERIOD_1MS - 1;
    SysTick->CTRL = SysTick_CTRL_ENABLE_Msk | SysTick_CTRL_TICKINT_Msk | SysTick_CTRL_CLKSOURCE_Msk;

    // SD card identification and FS test - cyclic and infinite
    While_TestFS();
}

void While_TestFS(void)
{
    volatile SD_Detect_Result_TypeDef CardDetectResult;

    while (1) {
        // LED_CYCLE blinks to show MCU is working
        if (!TmrDelay_ms) {
            PORT_WriteBit(MDR_PORTB, LED_CYCLE, (BitStatus)!PORT_ReadInputDataBit(MDR_PORTB, LED_CYCLE));
            TmrDelay_ms = 1000;
        }

        //  File System Test
        if (!PORT_ReadInputDataBit(MDR_PORTC, PORT_Pin_4)) {
            while (!PORT_ReadInputDataBit(MDR_PORTC, PORT_Pin_4)) { }

            PORT_WriteBit(MDR_PORTB, LED_SEL_ALL, RESET);

            if (TestFileSystem() == 0)
                PORT_WriteBit(MDR_PORTB, LED_OK, SET);
            else
                PORT_WriteBit(MDR_PORTB, LED_FAULT, SET);
        }

        //  Card identification test
        if (!PORT_ReadInputDataBit(MDR_PORTC, PORT_Pin_0)) {
            while (!PORT_ReadInputDataBit(MDR_PORTC, PORT_Pin_0)) { }

            PORT_WriteBit(MDR_PORTB, LED_SEL_ALL, RESET);

            CardDetectResult = SD_CARD_CardDetect();
            if (CardDetectResult == SD_OK) {
                printf("SDIO Card Detected!\n");
                PORT_WriteBit(MDR_PORTB, LED_OK, SET);
            } else {
                printf("SDIO Card Detection FAILED!\n");
                PORT_WriteBit(MDR_PORTB, LED_FAULT, SET);
            }
        }
    }
}

// SysTick handelr
void SysTick_Handler(void)
{
    if (TmrDelay_ms)
        --TmrDelay_ms;

    SDIO_LL_DelayHandler();
}

