
// High-level SD card protocol abstraction

#include "sd_card_protocol.h"
#include "sd_card_sdio_ll.h"

#define SD_DriverInit       SDIO_LL_Init

#define SD_PowerOn          SDIO_LL_PowerOn
#define SD_PowerOff         SDIO_LL_PowerOff
#define SD_SetRateBRG       SDIO_LL_SetClockBRG

#define SD_ExecCMD          SDIO_LL_ExecCMD

#define SD_WriteBlock       SDIO_LL_WriteBlock
#define SD_ReadBlock        SDIO_LL_ReadBlock
#define SD_Send_ClockCycles SDIO_LL_SendClockCycles
#define SD_StopTransfer     SDIO_LL_StopTransfer

#define SD_GetPowerStatus   SDIO_LL_GetPowerStatus
#define SD_IsCardInsert     SDIO_LL_IsCardInsert

#define SD_RunTimeout       SDIO_LL_RunTimeout
#define SD_GetTimeout       SDIO_LL_GetTimeout

static SD_CARD_ClockCfg_TypeDef SD_ClockConfig;

static uint8_t  CardType, CardInfo[16 + 16 + 4]; // Card type flag, CSD(16), CID(16), OCR(4)
static uint32_t CardRCA = 0;                     // Relative Card Address
static uint32_t CardReady;

//  ============= Initialization ====================

void SD_CARD_Init(SD_CARD_ClockCfg_TypeDef* cfgClock, SD_CARD_Timeouts_TypeDef* cfgTimeouts)
{
    // Initialize SDIO
    SD_ClockConfig = *cfgClock;
    SD_DriverInit(SD_ClockConfig.SysTimerPeriod_1ms, cfgTimeouts);
}

//  ============= Auxilary funcions ====================
static void bswap_cp(uint8_t* dst, const uint32_t* src)
{
    uint32_t d = *src;

    *dst++ = (uint8_t)(d >> 24);
    *dst++ = (uint8_t)(d >> 16);
    *dst++ = (uint8_t)(d >> 8);
    *dst++ = (uint8_t)(d >> 0);
}

// Unified function for CMD and ACMD transfers
bool SD_ExecACMD(uint32_t cmd, uint32_t arg, SD_CARD_ResponseLength_TypeDef respLength, uint32_t* response)
{
    SD_ResponseR1              respCMD55;
    SD_CARD_CMD_Result_TypeDef result;

    if (cmd & 0x80) {
        // Switch SD card to ACMD command response waiting
        result = SD_ExecCMD(CMD55, CardRCA << 16, SD_CARD_Resp_48bit, &respCMD55.value);
        if (result != SD_CARD_CMD_OK)
            return false;

        // Exit if SD card doesnt return APP_CMD wait status
        if (!(respCMD55.bits.SD_R1_APP_CMD))
            return false;
    }

    //  Command "Application-Specific Command"
    return (SD_ExecCMD(cmd, arg, respLength, response) == SD_CARD_CMD_OK);
}

bool SD_CARD_WaitForCardStatus(uint32_t statusMask, uint32_t timeout_ms)
{
    uint32_t rc;

    SD_RunTimeout(timeout_ms);
    while (SD_GetTimeout()) {
        //  CMD13 - SEND_STATUS
        if (SD_ExecACMD(CMD13, CardRCA << 16, SD_CARD_Resp_48bit, &rc) && ((rc & statusMask) == statusMask))
            return true;
    }
    return false;
}

bool SD_CARD_IsCardReady(void)
{
    return CardReady;
}

bool SD_CARD_IsCardInsert(void)
{
    return SD_IsCardInsert();
}

void SD_Delay_ms(uint32_t delay_ms)
{
    SD_RunTimeout(delay_ms);
    while (SD_GetTimeout())
        ;
}

//  ============= SD Card Detect ====================

SD_Detect_Result_TypeDef SD_DetectCard(void)
{
    uint32_t          cmd, n;
    uint32_t          resp[4];
    SD_ResponseR3     respR3;
    SD_ResponseR1     respR1;
    volatile uint32_t ty = 0;

    // SD card command to enter IDLE state
    SD_ExecACMD(CMD0, 0, SD_CARD_Resp_No, resp); // CMD0 - GO_IDLE_STATE

    CardRCA = 0;

    //  Interface support request:
    //    [11..8] - Supply voltage,
    //              0001b - [2.7V .. 3.6V]
    //              0010b - LowVoltage (reserved)
    //    [7..0]  - check mask, 0x1AA
    SD_RunTimeout(1000);
    if (SD_ExecACMD(CMD8, 0x1AA, SD_CARD_Resp_48bit, resp)) // CMD8 - SEND_IF_COND
    {                                                       // SD card Ver2.00 or newer

        // Response R7
        // Returns mask if supports defined voltage
        if ((resp[0] & 0xFFF) != 0x1AA) { // Doesnt support supply 2.7-3.6V
            return SD_VER2_RespErr;
        } else { // SD card supports supply voltage 2.7-3.6V

            //  Request to send OCR register - Operation Conditions Register:
            //    [0..23] - Supply voltage window
            //       [30] - HCS (Host Capacity Support, bit30) - Say SD card to support  Tell SD card that a high capacity is maintained.
            //  Response R3 - wait for setting most significant bit SD_R3_Ready == 1 - initialization after PowerUp is done.
            while (!SD_ExecACMD(ACMD41, 0x40FF8000, SD_CARD_Resp_48bit, &respR3.value) || !(respR3.bits.SD_R3_Ready)) // ACMD41 - SEND_OP_COND (SDC)
            {
                if (!SD_GetTimeout())
                    return SD_VER2_READY_OV;
            }

            //  Check CCS (Card Capacity Status) bit in OCR register.
            //    CCS = 1 - High Capacity SD Memory
            //    CCS = 0 - Standard Capacity SD Memory Card
            ty = respR3.bits.SD_R3_CCS ? CT_SD2 | CT_BLOCK : CT_SD2;
        }
    } else { // SD card Ver1 or MMC
        //  Request to send OCR register - Operation Conditions Register:
        //    [0..23] - Supply voltage window
        if (SD_ExecACMD(ACMD41, 0x00FF8000, SD_CARD_Resp_48bit, resp)) // ACMD41 - SEND_OP_COND (SDC)
        {
            ty  = CT_SD1; // ACMD41 is accepted -> SDC Ver1
            cmd = ACMD41; // ACMD41 - SEND_OP_COND (SDC)
        } else {
            ty  = CT_MMC; // ACMD41 is rejected -> MMC
            cmd = CMD1;   // CMD1 - SEND_OP_COND (MMC)
        }

        //  Response R3 - wait for setting most significant bit SD_R3_Ready == 1 - initialization after PowerUp is done.
        while (!SD_ExecACMD(cmd, 0x00FF8000, SD_CARD_Resp_48bit, &respR3.value) || !(respR3.bits.SD_R3_Ready)) {
            if (!SD_GetTimeout())
                return SD_VER1_READY_OV;
        };
    }

    // Card is initialized, save card type and OCR register
    CardType = ty;
    bswap_cp(&CardInfo[32], resp);

    // ---------  Card in READY state  ------------------

    // Get CID register - Card Identification Data
    //   CMD2 - ALL_SEND_CID
    if (!SD_ExecACMD(CMD2, 0, SD_CARD_Resp_136bit, resp))
        return SD_CMD2_NO_RESP;
    for (n = 0; n < 4; n++)
        bswap_cp(&CardInfo[n * 4 + 16], &resp[n]); // Save CID

    // Get relative address RCA - Relative Card Address
    //   CMD3 - SEND_RELATIVE_ADDR
    if (ty & CT_SDC) {
        if (!SD_ExecACMD(CMD3, 0, SD_CARD_Resp_48bit, resp)) // SD Card
            return SD_CMD3_SD_NO_RESP;
        CardRCA = (resp[0] >> 16);
    } else {
        if (!SD_ExecACMD(CMD3, 1 << 16, SD_CARD_Resp_48bit, resp)) // MMC Card
            return SD_CMD3_MMC_NO_RESP;
        CardRCA = 1;
    }

    // Get CSD register - Card Specific Data
    //   CMD9 - SEND_CSD
    if (!SD_ExecACMD(CMD9, CardRCA << 16, SD_CARD_Resp_136bit, resp))
        return SD_CMD9_NO_RESP;
    for (n = 0; n < 4; n++)
        bswap_cp(&CardInfo[n * 4], &resp[n]);

    // Switch card to Transfer state
    // Only one card on bus could be in Transfer state (selected as active)
    // Other cards are in Standby mode. Activation of new card will switch current card to Standby mode automatically
    //    CMD7  - SELECT_CARD
    // Response - R1b, identical to the R1 response with an optional busy signal transmitted over the card output data line.
    // The card may be busy with an internal operation after receiving commands, depending on its state before receiving the command.
    // Host needs to check the card business on every card answer - NOT IMPLEMENTED
    if (!SD_ExecACMD(CMD7, CardRCA << 16, SD_CARD_Resp_48bit, resp))
        return SD_CMD7_NO_RESP;

    // ---------  Card in TRANSFER State  ------------------

    // Changing the 512 byte block size for Standard Capacity cards
    //   CMD16 - SET_BLOCKLEN
    if (!(ty & CT_BLOCK)) {
        if (!SD_ExecACMD(CMD16, SD_BLOCK_512, SD_CARD_Resp_48bit, &respR1.value) || (respR1.value & SD_RespR1_All_ErrorsMask))
            return SD_CMD16_NO_RESP;
    }

    if (ty & CT_SDC) {
        // Set the data bus to 4 bits, (after IDLE width = 1 bit)
        //   ACMD6 - SET_BUS_WIDTH (SDC)
        if (!SD_ExecACMD(ACMD6, 2, SD_CARD_Resp_48bit, &respR1.value) || (respR1.value & SD_RespR1_All_ErrorsMask))
            return SD_ACMD6_NO_RESP;
    }

    return SD_OK;
}

SD_Detect_Result_TypeDef SD_CARD_CardDetect(void)
{
    SD_Detect_Result_TypeDef result;

    CardReady = false;

    SD_PowerOn();
    // Frequency during card initialization 100KHz .. 400KHz
    SD_SetRateBRG(SD_ClockConfig.SD_CLK_BRG_Detect);
    // The power-on-delay, time of 74 clock cycles
//    SD_Delay_ms(SD_ClockConfig.SD_PowerOn_Delay_ms);

    // Sending 74 clock cycles is not necessary, works without it with power-on-delay
//    SD_Send_ClockCycles(32);
//    SD_Send_ClockCycles(32);
//    SD_Send_ClockCycles(10);

    // Pooling card
    result = SD_DetectCard();

    // Set work exchange rate
    if (result == SD_OK) {
        SD_SetRateBRG(SD_ClockConfig.SD_CLK_BRG_Transfer);
        CardReady = true;
    } else
        SD_PowerOff();

    return result;
}

//  ============= SD Read Data ====================

SD_RW_Result_TypeDef SD_CARD_CardRead(uint8_t* buff, uint32_t sectorNum, uint8_t sectorCnt)
{
    uint32_t resp;
    uint32_t cmd;

    if (sectorCnt < 1 || sectorCnt > 127)
        return SD_RW_PARERR;
    if (!CardReady)
        return SD_RW_NOTRDY;

    if (!(CardType & CT_BLOCK))
        sectorNum *= SD_BLOCK_512;
    if (!SD_CARD_WaitForCardStatus(SD_State_Msk_Transfer, SD_ClockConfig.SD_WaitStatus_Timout_ms))
        return SD_RW_ERROR;

    //  CMD17 - READ_SINGLE_BLOCK
    //  CMD18 - READ_MULTIPLE_BLOCK
    cmd = (sectorCnt > 1) ? CMD18 : CMD17;
    if (SD_ExecACMD(cmd, sectorNum, SD_CARD_Resp_48bit, &resp) && !(resp & SD_RespR1_RW_ErrorsMask)) {
        do {
            if (!SD_ReadBlock(buff, SD_BLOCK_512))
                break;
            buff += SD_BLOCK_512;
        } while (--sectorCnt);

        // Stop reading MULTIPLE_BLOCK
        //   CMD12 - STOP_TRANSMISSION
        if (cmd == CMD18)
            SD_ExecACMD(CMD12, 0, SD_CARD_Resp_48bit, &resp);
    }

    SD_Send_ClockCycles(8);

    return sectorCnt ? SD_RW_ERROR : SD_RW_OK;
}

//  ============= SD Write Data ====================

bool SD_TryWriteBlock(const uint8_t* buff)
{
    if (!SD_CARD_WaitForCardStatus(SD_State_Msk_Receive | SD_State_Msk_ReadyForData, SD_ClockConfig.SD_WaitStatus_Timout_ms))
        return false;

    return SD_WriteBlock(buff);
}

SD_RW_Result_TypeDef SD_CARD_CardWrite(const uint8_t* buff, uint32_t sectorNum, uint8_t sectorCnt)
{
    uint32_t rc;
    uint32_t cmd;

    if (sectorCnt < 1 || sectorCnt > 127)
        return SD_RW_PARERR;
    if (!CardReady)
        return SD_RW_NOTRDY;

    if (!(CardType & CT_BLOCK))
        sectorNum *= SD_BLOCK_512;
    if (!SD_CARD_WaitForCardStatus(SD_State_Msk_Transfer, SD_ClockConfig.SD_WaitStatus_Timout_ms))
        return SD_RW_ERROR;

    if (sectorCnt == 1)
        cmd = CMD24; // CMD24 - WRITE_BLOCK
    else {
        //  CMD23  - SET_BLK_COUNT (MMC)
        //  ACMD23 - SET_WR_BLK_ERASE_COUNT (SDC)
        cmd = (CardType & CT_SDC) ? ACMD23 : CMD23;
        if (!SD_ExecACMD(cmd, sectorCnt, SD_CARD_Resp_48bit, &rc) || (rc & SD_RespR1_RW_ErrorsMask)) {
            return SD_RW_ERROR;
        }

        cmd = CMD25; // CMD25 - WRITE_MULTIPLE_BLOCK
    }

    // Send a write command
    if (!SD_ExecACMD(cmd, sectorNum, SD_CARD_Resp_48bit, &rc) || (rc & SD_RespR1_RW_ErrorsMask))
        return SD_RW_ERROR;

    while (sectorCnt) {
        if (!SD_TryWriteBlock(buff))
            break;
        sectorCnt--;
        buff += SD_BLOCK_512;
    }

    // Stop reading MULTIPLE_BLOCK
    //   CMD12 - STOP_TRANSMISSION
    if (cmd == CMD25 && (CardType & CT_SDC))
        SD_ExecACMD(CMD12, 0, SD_CARD_Resp_48bit, &rc);

    SD_Send_ClockCycles(8);

    return sectorCnt ? SD_RW_ERROR : SD_RW_OK;
}

//  ============= SD Misc Control ====================
SD_RW_Result_TypeDef SD_CARD_IO_Ctrl(SD_Card_IO_Ctrl_TypeDef ctrl, void* buff)
{
    SD_RW_Result_TypeDef res;
    uint8_t              b, *ptr = buff;
    uint32_t             resp[4], d, *dp, st, ed;

    res = SD_RW_ERROR;

    switch (ctrl) {
        case SDC_CTRL_SYNC:                                                                               // Make sure that all data has been written on the media
            if (SD_CARD_WaitForCardStatus(SD_State_Msk_Transfer, SD_ClockConfig.SD_WaitStatus_Timout_ms)) // Wait for card enters tran state
                res = SD_RW_OK;
            break;

        case SDC_GET_SECTOR_COUNT: // Get number of sectors on the disk (DWORD)
            if ((CardInfo[0] >> 6) == 1) {
                // SDC CSD v2.0
                d                = ((uint16_t)CardInfo[8] << 8) + CardInfo[9] + 1;
                *(uint32_t*)buff = d << 10;
            } else {
                // MMC or SDC CSD v1.0
                b                = (CardInfo[5] & 15) + ((CardInfo[10] & 128) >> 7) + ((CardInfo[9] & 3) << 1) + 2;
                d                = (CardInfo[8] >> 6) + ((uint16_t)CardInfo[7] << 2) + ((uint16_t)(CardInfo[6] & 3) << 10) + 1;
                *(uint32_t*)buff = d << (b - 9);
            }
            res = SD_RW_OK;
            break;

        case SDC_GET_BLOCK_SIZE: // Get erase block size in unit of sectors (DWORD)
            if (CardType & CT_SD2) {
                // SDC ver 2.00
                *(uint32_t*)buff = 16UL << (CardInfo[10] >> 4);
            } else {
                // SDC ver 1.XX or MMC
                if (CardType & CT_SD1) // SDC v1
                    *(uint32_t*)buff = (((CardInfo[10] & 63) << 1) + ((uint16_t)(CardInfo[11] & 128) >> 7) + 1) << ((CardInfo[13] >> 6) - 1);
                else // MMC
                    *(uint32_t*)buff = ((uint16_t)((CardInfo[10] & 124) >> 2) + 1) * (((CardInfo[11] & 3) << 3) + ((CardInfo[11] & 224) >> 5) + 1);
            }
            res = SD_RW_OK;
            break;

        case SDC_CTRL_ERASE_SECTOR: // Erase a block of sectors
            // Check if sector erase can be applied to the card
            if (!(CardType & CT_SDC) || (!(CardInfo[0] >> 6) && !(CardInfo[10] & 0x40)))
                break;

            dp = buff;
            st = dp[0];
            ed = dp[1];

            if (!(CardType & CT_BLOCK)) {
                st *= SD_BLOCK_512;
                ed *= SD_BLOCK_512;
            }

            if (SD_ExecACMD(CMD32, st, SD_CARD_Resp_48bit, resp) && SD_ExecACMD(CMD33, ed, SD_CARD_Resp_48bit, resp) &&
                SD_ExecACMD(CMD38, 0, SD_CARD_Resp_48bit, resp) && SD_CARD_WaitForCardStatus(SD_State_Msk_Transfer, SD_ClockConfig.SD_Erase_Timout_ms))
                res = SD_RW_OK;
            break;

        case SDC_CTRL_POWER:
            switch (ptr[0]) {
                case 0:
                    SD_PowerOff();
                    res = SD_RW_OK;
                    break;
                case 1:
                    ptr[1] = (uint8_t)SD_GetPowerStatus();
                    res    = SD_RW_OK;
                    break;
                default:
                    res = SD_RW_PARERR;
            }
            break;

        case SDC_MMC_GET_TYPE: // Get card type flags (1 byte)
            *ptr = CardType;
            res  = SD_RW_OK;
            break;

        case SDC_MMC_GET_CSD: // Get CSD (16 bytes)
            memcpy(buff, &CardInfo[0], 16);
            res = SD_RW_OK;
            break;

        case SDC_MMC_GET_CID: // Get CID (16 bytes)
            memcpy(buff, &CardInfo[16], 16);
            res = SD_RW_OK;
            break;

        case SDC_MMC_GET_OCR: // Get OCR (4 bytes)
            memcpy(buff, &CardInfo[32], 4);
            res = SD_RW_OK;
            break;

        case SDC_MMC_GET_SDSTAT: // Receive SD status as a data block (64 bytes)
            if (CardType & CT_SDC) {
                // SDC
                if (SD_CARD_WaitForCardStatus(SD_State_Msk_Transfer, SD_ClockConfig.SD_WaitStatus_Timout_ms)) {
                    if (SD_ExecACMD(ACMD13, 0, SD_CARD_Resp_48bit, resp) // Start to read
                        && !(resp[0] & SD_RespR1_RW_ErrorsMask)) {
                        if (SD_ReadBlock(buff, SD_BLOCK_512))
                            res = SD_RW_OK;
                    }
                }
                SD_StopTransfer();
            }
            break;

        default:
            res = SD_RW_PARERR;
    }

    return res;
}
