
// High-level SD card protocol abstraction

#ifndef SDIO_SD_CARD_PROTOCOL_H
#define SDIO_SD_CARD_PROTOCOL_H

#include <stdbool.h>
#include <stdint.h>

// Response token size
typedef enum {
    SD_CARD_Resp_No     = 0,
    SD_CARD_Resp_48bit  = 48,
    SD_CARD_Resp_136bit = 136,
} SD_CARD_ResponseLength_TypeDef;

#define SD_CMD_LEN   48
#define SD_BLOCK_512 512

// Status of execution of the command to the CD card
typedef enum {
    SD_CARD_CMD_OK,
    SD_CARD_CMD_TimeoutSend,
    SD_CARD_CMD_TimeoutRead,
} SD_CARD_CMD_Result_TypeDef;

// Time delays when working with the card
typedef struct
{
    uint32_t SD_SendCMD_Timout_ms;    //  100 ms
    uint32_t SD_BeforeRead_Delay_ms;  //    2 ms
    uint32_t SD_ReadResp_Timout_ms;   //  100 ms
    uint32_t SD_ChangeRate_Timout_ms; //    2 ms
    uint32_t SD_ReadBlock_Timout_ms;  //  100 ms
    uint32_t SD_WriteBlock_Timout_ms; // 1000 ms
} SD_CARD_Timeouts_TypeDef;

// The parameters depend on the current clock frequency

// Values for SD_CLK_BRG_Detect and SD_CLK_BRG_Transfer
#define SDIO_BRG_div4   1
#define SDIO_BRG_div8   2
#define SDIO_BRG_div16  3
#define SDIO_BRG_div32  4
#define SDIO_BRG_div64  5
#define SDIO_BRG_div128 6
#define SDIO_BRG_div256 7

typedef struct
{
    uint32_t SysTimerPeriod_1ms;  // Value to get 1ms timer period
    uint32_t SD_CLK_BRG_Detect;   // HCLK/BRG = CLK_freq    100KHz.. 400KHz
    uint32_t SD_CLK_BRG_Transfer; // HCLK/BRG = CLK_freq < 25MHz

    uint32_t SD_PowerOn_Delay_ms;     //     2 ms
    uint32_t SD_WaitStatus_Timout_ms; //   500 ms
    uint32_t SD_Erase_Timout_ms;      // 30000 ms
} SD_CARD_ClockCfg_TypeDef;

/* ----- MMC/SDC command ----- */
#define CMD0   (0)         // GO_IDLE_STATE
#define CMD1   (1)         // SEND_OP_COND (MMC)
#define CMD2   (2)         // ALL_SEND_CID
#define CMD3   (3)         // SEND_RELATIVE_ADDR
#define ACMD6  (6 | 0x80)  // SET_BUS_WIDTH (SDC)
#define CMD7   (7)         // SELECT_CARD
#define CMD8   (8)         // SEND_IF_COND
#define CMD9   (9)         // SEND_CSD
#define CMD10  (10)        // SEND_CID
#define CMD12  (12)        // STOP_TRANSMISSION
#define CMD13  (13)        // SEND_STATUS
#define ACMD13 (13 | 0x80) // SD_STATUS (SDC)
#define CMD16  (16)        // SET_BLOCKLEN
#define CMD17  (17)        // READ_SINGLE_BLOCK
#define CMD18  (18)        // READ_MULTIPLE_BLOCK
#define CMD23  (23)        // SET_BLK_COUNT (MMC)
#define ACMD23 (23 | 0x80) // SET_WR_BLK_ERASE_COUNT (SDC)
#define CMD24  (24)        // WRITE_BLOCK
#define CMD25  (25)        // WRITE_MULTIPLE_BLOCK
#define CMD32  (32)        // ERASE_ER_BLK_START
#define CMD33  (33)        // ERASE_ER_BLK_END
#define CMD38  (38)        // ERASE
#define ACMD41 (41 | 0x80) // SEND_OP_COND (SDC)
#define CMD55  (55)        // APP_CMD

/* Card type flags (CardType) */
#define CT_MMC   0x01              /* MMC ver 3 */
#define CT_SD1   0x02              /* SD ver 1 */
#define CT_SD2   0x04              /* SD ver 2 */
#define CT_SDC   (CT_SD1 | CT_SD2) /* SD */
#define CT_BLOCK 0x08              /* Block addressing */

// ------------------  Command code for disk_ioctrl function --------------
typedef enum {
    /* Generic ioctl command (defined for FatFs) */
    SDC_CTRL_SYNC         = 0, /* Flush disk cache (for write functions) */
    SDC_GET_SECTOR_COUNT  = 1, /* Get media size (for only f_mkfs()) */
    SDC_GET_SECTOR_SIZE   = 2, /* Get sector size (for multiple sector size (_MAX_SS >= 1024)) */
    SDC_GET_BLOCK_SIZE    = 3, /* Get erase block size (for only f_mkfs()) */
    SDC_CTRL_ERASE_SECTOR = 4, /* Force erased a block of sectors (for only _USE_ERASE) */

    /* Generic ioctl command */
    SDC_CTRL_POWER = 5, /* Get/Set power status */
    SDC_CTRL_LOCK  = 6, /* Lock/Unlock media removal */
    SDC_CTRL_EJECT = 7, /* Eject media */

    /* MMC/SDC specific ioctl command */
    SDC_MMC_GET_TYPE   = 10, /* Get card type */
    SDC_MMC_GET_CSD    = 11, /* Get CSD */
    SDC_MMC_GET_CID    = 12, /* Get CID */
    SDC_MMC_GET_OCR    = 13, /* Get OCR */
    SDC_MMC_GET_SDSTAT = 14, /* Get SD status */

    /* ATA/CF specific ioctl command */
    SDC_ATA_GET_REV   = 20, /* Get F/W revision */
    SDC_ATA_GET_MODEL = 21, /* Get model name */
    SDC_ATA_GET_SN    = 22, /* Get serial number */

    /* NAND specific ioctl command */
    SDC_NAND_FORMAT = 30 /* Create physical format */
} SD_Card_IO_Ctrl_TypeDef;

//=============   Card Responses ===================
typedef enum {
    SD_State_Idle           = 0,
    SD_State_Ready          = 1,
    SD_State_Identification = 2,
    SD_State_Standby        = 3,
    SD_State_Transfer       = 4,
    SD_State_DataSending    = 5,
    SD_State_Receive        = 6,
    SD_State_Programming    = 7,
    SD_State_Disconnect     = 8,
    SD_State_Reserved1      = 9,
    SD_State_Reserved2      = 10,
    SD_State_Reserved3      = 11,
    SD_State_Reserved4      = 12,
    SD_State_Reserved5      = 13,
    SD_State_Reserved6      = 14,
    SD_State_Reserved7      = 15,
} SD_State_TypeDef;

#define SD_State_Offs_Status        9
#define SD_State_Msk_Idle           ((uint32_t)(SD_State_Idle << SD_State_Offs_Status))
#define SD_State_Msk_Ready          ((uint32_t)(SD_State_Ready << SD_State_Offs_Status))
#define SD_State_Msk_Identification ((uint32_t)(SD_State_Identification << SD_State_Offs_Status))
#define SD_State_Msk_Standby        ((uint32_t)(SD_State_Standby << SD_State_Offs_Status))
#define SD_State_Msk_Transfer       ((uint32_t)(SD_State_Transfer << SD_State_Offs_Status))
#define SD_State_Msk_DataSending    ((uint32_t)(SD_State_DataSending << SD_State_Offs_Status))
#define SD_State_Msk_Receive        ((uint32_t)(SD_State_Receive << SD_State_Offs_Status))
#define SD_State_Msk_Programming    ((uint32_t)(SD_State_Programming << SD_State_Offs_Status))
#define SD_State_Msk_Disconnect     ((uint32_t)(SD_State_Disconnect << SD_State_Offs_Status))

#define SD_State_Offs_ReadyForData  8
#define SD_State_Msk_ReadyForData   ((uint32_t)(1 << SD_State_Offs_ReadyForData))

typedef struct
{
    uint32_t         SD_R1_Reserv1         : 3; //  0..2 bits
    bool             SD_R1_AKE_SEQ_ERR     : 1; //  3 bit
    uint32_t         SD_R1_Reserv2         : 1; //  4 bit
    bool             SD_R1_APP_CMD         : 1; //  5 bit
    uint32_t         SD_R1_Reserv3         : 2; //  6..7 bits
    bool             SD_R1_ReadyForData    : 1; //  8 bit
    SD_State_TypeDef SD_R1_State           : 4; //  9..12 bit
    bool             SD_R1_EraseReset      : 1; //  13 bit
    bool             SD_R1_ECCDisable      : 1; //  14 bit
    bool             SD_R1_WPEraseSkip     : 1; //  15 bit
    bool             SD_R1_CSDOverwriteErr : 1; //  16 bit
    uint32_t         SD_R1_Reserv4         : 2; //  17..18 bits
    bool             SD_R1_Error           : 1; //  19 bit
    bool             SD_R1_CCError         : 1; //  20 bit
    bool             SD_R1_ECCFailed       : 1; //  21 bit
    bool             SD_R1_IllCommand      : 1; //  22 bit
    bool             SD_R1_CRCErr          : 1; //  23 bit
    bool             SD_R1_LockFailed      : 1; //  24 bit
    bool             SD_R1_IsLocked        : 1; //  25 bit
    bool             SD_R1_WPViolation     : 1; //  26 bit
    bool             SD_R1_EraseParam      : 1; //  27 bit
    bool             SD_R1_EraseSeqErr     : 1; //  28 bit
    bool             SD_R1_BlockLenErr     : 1; //  29 bit
    bool             SD_R1_AddrErr         : 1; //  30 bit
    bool             SD_R1_OutOfRange      : 1; //  31 bit
} SD_RespR1_Bits;

typedef union {
    uint32_t       value;
    SD_RespR1_Bits bits;
} SD_ResponseR1;

#define SD_RespR1_All_ErrorsMask 0xFDF90000
#define SD_RespR1_RW_ErrorsMask  0xC0580000

typedef struct
{
    uint32_t SD_R3_Reserv1 : 15; //  0..14 bits
    bool     SD_R3_2V7_2V8 : 1;  //  15 bit
    bool     SD_R3_2V8_2V9 : 1;  //  16 bit
    bool     SD_R3_2V9_3V0 : 1;  //  17 bit
    bool     SD_R3_3V0_3V1 : 1;  //  18 bit
    bool     SD_R3_3V1_3V2 : 1;  //  19 bit
    bool     SD_R3_3V2_3V3 : 1;  //  20 bit
    bool     SD_R3_3V3_3V4 : 1;  //  21 bit
    bool     SD_R3_3V4_3V5 : 1;  //  22 bit
    bool     SD_R3_3V5_3V6 : 1;  //  23 bit
    uint32_t SD_R3_Reserv2 : 6;  //  24..29 bits
    bool     SD_R3_CCS     : 1;  //  30 bit  - Card Capacity Status
    bool     SD_R3_Ready   : 1;  //  31 bit  - Ready after initialization
} SD_RespR3_Bits;

typedef union {
    uint32_t       value;
    SD_RespR3_Bits bits;
} SD_ResponseR3;

typedef enum {
    SD_OK,
    SD_VER2_RespErr, //  Power response is wrong
    SD_VER2_READY_OV,
    SD_VER1_READY_OV,
    SD_CMD2_NO_RESP,     //  CID Read fault
    SD_CMD3_SD_NO_RESP,  //  RCA Read fault - SD
    SD_CMD3_MMC_NO_RESP, //  RCA Read fault - MMS
    SD_CMD9_NO_RESP,     //  CSD Read fault (Card Specific Data)
    SD_CMD7_NO_RESP,     //  SELECT_CARD fault
    SD_CMD16_NO_RESP,    //  SET_BLOCKLEN 512 fault
    SD_ACMD6_NO_RESP     //  SET_BUS_WIDTH = 4 fault
} SD_Detect_Result_TypeDef;

typedef enum {
    SD_RW_OK = 0, // 0: Successful
    SD_RW_ERROR,  // 1: R/W Error
    SD_RW_WRPRT,  // 2: Write Protected
    SD_RW_NOTRDY, // 3: Not Ready
    SD_RW_PARERR  // 4: Invalid Parameter
} SD_RW_Result_TypeDef;

// Setting the operating parameters
void SD_CARD_Init(SD_CARD_ClockCfg_TypeDef* cfgClock, SD_CARD_Timeouts_TypeDef* cfgTimeouts);

// Work with SD card
SD_Detect_Result_TypeDef SD_CARD_CardDetect(void);
SD_RW_Result_TypeDef     SD_CARD_CardRead(uint8_t* buff, uint32_t sectorNum, uint8_t sectorCnt);
SD_RW_Result_TypeDef     SD_CARD_CardWrite(const uint8_t* buff, uint32_t sectorNum, uint8_t sectorCnt);
SD_RW_Result_TypeDef     SD_CARD_IO_Ctrl(SD_Card_IO_Ctrl_TypeDef ctrl, void* buff);
bool                     SD_CARD_IsCardReady(void);
bool                     SD_CARD_IsCardInsert(void);

// For debug purpose
bool SD_CARD_WaitForCardStatus(uint32_t statusMask, uint32_t timeout_ms);

#endif // SDIO_SD_CARD_PROTOCOL_H
