
// API for low-level SD card protocol implementation

#include "sd_card_sdio_ll.h"
#include "math.h"

#define BRD_SDIO_PIN_CLK       PORT_Pin_8  // PC8
#define BRD_SDIO_PIN_CMD       PORT_Pin_1  // PC1
#define BRD_SDIO_PIN_D0        PORT_Pin_10 // PC10
#define BRD_SDIO_PIN_D1        PORT_Pin_11 // PC11
#define BRD_SDIO_PIN_D2        PORT_Pin_12 // PC12
#define BRD_SDIO_PIN_D3        PORT_Pin_13 // PC13

#define BRD_SDIO_PORT_CLK      RST_CLK_PCLK_PORTC
#define BRD_SDIO_PORT          MDR_PORTC

#define BRD_SDIO_PORT_D0       MDR_PORTC //  for Write ack

#define BRD_SDIO_INIT_PORT     MDR_PORTC
#define BRD_SDIO_INIT_PIN      BRD_SDIO_PIN_D0 | BRD_SDIO_PIN_D1 | BRD_SDIO_PIN_D2 | BRD_SDIO_PIN_D3
#define BRD_SDIO_INIT_FUNC     PORT_FUNC_MAIN
#define BRD_SDIO_INIT_PULLUP   PORT_PULL_UP_ON

#define BRD_SDIO_INIT_PORT_1   MDR_PORTC
#define BRD_SDIO_INIT_PIN_1    BRD_SDIO_PIN_CMD
#define BRD_SDIO_INIT_FUNC_1   PORT_FUNC_OVERRID
#define BRD_SDIO_INIT_PULLUP_1 PORT_PULL_UP_ON

#define BRD_SDIO_INIT_PORT_2   MDR_PORTC
#define BRD_SDIO_INIT_PIN_2    BRD_SDIO_PIN_CLK
#define BRD_SDIO_INIT_FUNC_2   PORT_FUNC_MAIN
#define BRD_SDIO_INIT_PULLUP_2 PORT_PULL_UP_OFF

void SysTickStart(uint32_t ticks)
{
    SysTick->LOAD = ticks - 1;
    SysTick->CTRL = SysTick_CTRL_ENABLE_Msk | SysTick_CTRL_TICKINT_Msk | SysTick_CTRL_CLKSOURCE_Msk;
}

void SysTickStop(void)
{
    SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}

//  ==================  GPIO init ==================
#define PORT_SPEED_SEL PORT_SPEED_MAXFAST

void SDIO_PortInit(void)
{
    PORT_InitTypeDef GPIOInitStruct;

    //	Clock for PORT
    RST_CLK_PCLKcmd(BRD_SDIO_PORT_CLK, ENABLE);

    //  Common params
    GPIOInitStruct.PORT_PULL_DOWN = PORT_PULL_DOWN_OFF;
    GPIOInitStruct.PORT_PD_SHM    = PORT_PD_SHM_OFF;
    GPIOInitStruct.PORT_PD        = PORT_PD_DRIVER;
    GPIOInitStruct.PORT_GFEN      = PORT_GFEN_OFF;
    GPIOInitStruct.PORT_SPEED     = PORT_SPEED_SEL;
    GPIOInitStruct.PORT_MODE      = PORT_MODE_DIGITAL;

    //  Custom params
    GPIOInitStruct.PORT_Pin     = BRD_SDIO_INIT_PIN;
    GPIOInitStruct.PORT_PULL_UP = BRD_SDIO_INIT_PULLUP;
    GPIOInitStruct.PORT_FUNC    = BRD_SDIO_INIT_FUNC;

    PORT_Init(BRD_SDIO_INIT_PORT, &GPIOInitStruct);

#ifdef BRD_SDIO_INIT_PORT_1
    GPIOInitStruct.PORT_Pin     = BRD_SDIO_INIT_PIN_1;
    GPIOInitStruct.PORT_PULL_UP = BRD_SDIO_INIT_PULLUP_1;
    GPIOInitStruct.PORT_FUNC    = BRD_SDIO_INIT_FUNC_1;

    PORT_Init(BRD_SDIO_INIT_PORT_1, &GPIOInitStruct);
#endif

#ifdef BRD_SDIO_INIT_PORT_2
    GPIOInitStruct.PORT_Pin     = BRD_SDIO_INIT_PIN_2;
    GPIOInitStruct.PORT_PULL_UP = BRD_SDIO_INIT_PULLUP_2;
    GPIOInitStruct.PORT_FUNC    = BRD_SDIO_INIT_FUNC_2;

    PORT_Init(BRD_SDIO_INIT_PORT_2, &GPIOInitStruct);
#endif
}

//  ==================  Initialization and operating parameters ==================

//  Timeouts that are used when working with the SD card
static SD_CARD_Timeouts_TypeDef _SDIO_Timeouts;

void SDIO_LL_Init(uint32_t periodSysTimer_1ms, SD_CARD_Timeouts_TypeDef* cfgTimeouts)
{
    // Configure SDIO pins to SD card
    SDIO_PortInit();

    //  Clocking the SDIO block
    RST_CLK_PCLKcmd(RST_CLK_PCLK_SDIO, ENABLE);

    // Set the system timer period to 1 ms
    _SDIO_Timeouts = *cfgTimeouts;
    SysTickStart(periodSysTimer_1ms);
}

void SDIO_LL_DeInit()
{
    SysTickStop();

    SDIO_Cmd(DISABLE);
    RST_CLK_PCLKcmd(RST_CLK_PCLK_SDIO, DISABLE);
}

bool SDIO_LL_IsCardInsert(void)
{
    // Need to doublecheck if works correct
    return (BRD_SDIO_PORT_D0->RXTX & BRD_SDIO_PIN_D0) != 0;
}

//  ==================  Timeouts ==================
// By the SD_CARD_ClockCfg_TypeDef.SysTimerPeriod_1ms parameter it is necessary to set the system timer period to 1 ms.

volatile uint32_t DelayInt_ms, DelayExt_ms;

// The delay counters decrease by 1 ms, with each interrupt. The equality to 0 means the timeout is triggered.
void SDIO_LL_DelayHandler(void)
{
    volatile uint32_t n;

    n = DelayExt_ms;
    if (n)
        DelayExt_ms = --n;
    n = DelayInt_ms;
    if (n)
        DelayInt_ms = --n;
}

//  The "built-in" SysTick handler, to use if the project does not have its own.
//  If there is in-project handler, SDIO_LL_DelayHandler() should be called in it so that timeouts are counted.
#if USE_SYS_TIMER_HANDLER
void SysTick_Handler(void)
{
    SDIO_LL_DelayHandler();
}
#endif

//  --------- Timeouts for external use in sdio_sd_card_protocol.c ----------

void SDIO_LL_RunTimeout(uint32_t timeout_ms)
{
    DelayExt_ms = timeout_ms;
}

uint32_t SDIO_LL_GetTimeout(void)
{
    return DelayExt_ms;
}

//  ==================  Power supply control ==================
// Note that power supply is always enabled on demoboard
void SDIO_LL_PowerOn(void)
{
    MDR_SDIO->CR = SDIO_CR__PowerOn;
}

void SDIO_LL_PowerOff(void)
{
    MDR_SDIO->CR &= ~SDIO_CR_SDE;
}

bool SDIO_LL_GetPowerStatus(void)
{
    return (MDR_SDIO->CR & SDIO_CR_SDE) == 1;
}

//  ==================  Auxilary functions ==================

// Changing the speed of working with the SD card - the frequency of the CLK signal
void SDIO_LL_SetClockBRG(uint32_t clockBRG)
{
    uint32_t regCR = MDR_SDIO->CR;

    if (((regCR >> SDIO_CR_BR_Pos) & 0x7UL) == clockBRG)
        return;

    //  Change BRG
    regCR &= ~(7 << SDIO_CR_BR_Pos);
    MDR_SDIO->CR = regCR | (clockBRG << SDIO_CR_BR_Pos);

    //  Timeout
    if (_SDIO_Timeouts.SD_ChangeRate_Timout_ms > 0)
        for (DelayInt_ms = _SDIO_Timeouts.SD_ChangeRate_Timout_ms; DelayInt_ms;)
            ;
}

static inline void SDIO_CR_SetMode(uint32_t CR_Flags)
{
    uint32_t regCR = MDR_SDIO->CR;
    regCR &= ~(SDIO_CR__ModeClear);
    MDR_SDIO->CR = regCR | CR_Flags;
}

static inline bool WaitCMDCompleted(uint32_t timeout_ms)
{
    DelayInt_ms = timeout_ms;
    while ((MDR_SDIO->CR & SDIO_CR_WORK2) != 0) {
        // Timeout expired, exit with an error
        if (!DelayInt_ms)
            return false;
    }

    return true;
}

#define CHECK_CMD_STOPPED 0

#if CHECK_CMD_STOPPED

static inline void WaitCMDStopped(void)
{
    do {
        MDR_SDIO->CR &= ~(SDIO_CR_WORK2 | SDIO_CR_SBITCMD);
    } while (MDR_SDIO->CR & SDIO_CR_WORK2);
}

#define WAIT_CMD_STOPPED WaitCMDStopped
#else
#define WAIT_CMD_STOPPED __nop
#endif

// NOTE - max 32 cycles per funcion call
bool SDIO_LL_SendClockCycles(uint32_t clockCycles)
{
    WAIT_CMD_STOPPED();

    SDIO_CR_SetMode(SDIO_CR__SendCycles);
    MDR_SDIO->CMDDR        = 0x00FF;
    MDR_SDIO->CMD_TRANSFER = clockCycles; // Number of cycles
    MDR_SDIO->CMDCRC       = 0x00000000;

    MDR_SDIO->CR |= SDIO_CR_WORK2;

    if (!WaitCMDCompleted(_SDIO_Timeouts.SD_SendCMD_Timout_ms))
        return false;

    MDR_SDIO->CR &= ~SDIO_CR_DIRCMD; // Command RX
    return true;
}

void SDIO_LL_StopTransfer(void)
{
    MDR_SDIO->CR &= 0xFFFE07FF; // Disable SDIO interrupt
    MDR_SDIO->CR &= 0xFFFFFFF7; // Stop SDIO data transfer
}

//  ============= SD Exec Command ====================

SD_CARD_CMD_Result_TypeDef SDIO_LL_SendCMD(uint32_t cmd, uint32_t arg)
{
    SDIO_CR_SetMode(SDIO_CR__SendCMD);
    SDIO_SendCommand(cmd, arg);

    // The transfer is finished
    return SD_CARD_CMD_OK;
}

SD_CARD_CMD_Result_TypeDef SDIO_LL_ReadResponse(SD_CARD_ResponseLength_TypeDef respLength,
                                                uint32_t*                      response)
{
    uint32_t tmpBuff[5];

    // Read answer form SD card
    SDIO_CR_SetMode(SDIO_CR__ReadResp);

    MDR_SDIO->CMD_TRANSFER = (uint32_t)respLength;
    MDR_SDIO->CMDCRC       = 0x00000000;
    MDR_SDIO->CR |= SDIO_CR_WORK2;

    // Waiting for a response with a timeout
    if (!WaitCMDCompleted(_SDIO_Timeouts.SD_ReadResp_Timout_ms)) {
        // Force stopping a transaction
        MDR_SDIO->CR &= ~(SDIO_CR_WORK2 | SDIO_CR_SBITCMD);
        return SD_CARD_CMD_TimeoutSend;
    }

    // Reading received data from FIFO
    tmpBuff[0] = MDR_SDIO->CMDDR; // Read the response words 0
    tmpBuff[1] = MDR_SDIO->CMDDR; // Read the response words 1

    response[0] = ((tmpBuff[0] << 8) & 0xFFFFFF00) | ((tmpBuff[1] >> 8) & 0x000000FF);
    if (respLength == SD_CARD_Resp_136bit) {
        tmpBuff[2] = MDR_SDIO->CMDDR; // Read the response words 2
        tmpBuff[3] = MDR_SDIO->CMDDR; // Read the response words 3
        tmpBuff[4] = MDR_SDIO->CMDDR; // Read the response words 4

        response[1] = ((tmpBuff[1] << 8) & 0xFFFFFF00) | ((tmpBuff[2] >> 8) & 0x000000FF);
        response[2] = ((tmpBuff[2] << 8) & 0xFFFFFF00) | ((tmpBuff[3] >> 8) & 0x000000FF);
        response[3] = ((tmpBuff[3] << 8) & 0xFFFFFF00) | ((tmpBuff[4] >> 8) & 0x000000FF);
    }

    return SD_CARD_CMD_OK;
}

SD_CARD_CMD_Result_TypeDef SDIO_LL_ExecCMD(uint32_t                       cmd,
                                           uint32_t                       arg,
                                           SD_CARD_ResponseLength_TypeDef respLength,
                                           uint32_t*                      response)
{
    SD_CARD_CMD_Result_TypeDef result;

    result = SDIO_LL_SendCMD(cmd, arg);
    if (result == SD_CARD_CMD_OK) {
        if (respLength == SD_CARD_Resp_No)
            SDIO_LL_SendClockCycles(8);
        else {
            //  Delay after reset SDIO_SD_CR_WORK2 before read answer from SD card - 2ms
            if (_SDIO_Timeouts.SD_BeforeRead_Delay_ms > 0)
                for (DelayInt_ms = _SDIO_Timeouts.SD_BeforeRead_Delay_ms; DelayInt_ms;) { }

            //  Read card response
            result = SDIO_LL_ReadResponse(respLength, response);
        }
    }

    return result;
}

//  ============= SD Read Block ====================

#define READ_BY_ASM 1
__ASM void __get_DAT(volatile uint32_t* adr_datdr, uint8_t* adr_buff)
{
    // clang-format off
    ldr r0, [r0]
    rev r0, r0
    str r0, [r1]

    bx lr
    // clang-format on
}

bool SDIO_LL_ReadBlock(uint8_t* buff, uint32_t blockSize)
{
    uint32_t          i;
    volatile uint32_t tmp_DATDR;

    SDIO_CR_SetMode(SDIO_CR__ReadBlock);

    MDR_SDIO->DAT_TRANSFER = blockSize * 8 + (16 + 1); // Length
    MDR_SDIO->DAT0CRC      = 0x00000000;
    MDR_SDIO->DAT1CRC      = 0x00000000;
    MDR_SDIO->DAT2CRC      = 0x00000000;
    MDR_SDIO->DAT3CRC      = 0x00000000;
    MDR_SDIO->CR |= SDIO_CR_WORK1; //  Start

    i           = 0;
    DelayInt_ms = _SDIO_Timeouts.SD_ReadBlock_Timout_ms;
    while ((MDR_SDIO->CR & SDIO_CR_WORK1) != 0) {
        // Exit on timeout
        if (!DelayInt_ms) {
            MDR_SDIO->CR &= ~(SDIO_CR_WORK1 | SDIO_CR_SBITDAT);
            return false;
        }
        if (((MDR_SDIO->SR & 0x02) == 0) && (i < blockSize)) {
#ifdef READ_BY_ASM
            __get_DAT(&MDR_SDIO->DATDR, &buff[i]);
#else
            tmp_DATDR   = SDIO->SD_DATDR;
            buff[i]     = (BYTE)(tmp_DATDR >> 24 & 0x000000FF); // Read the response words
            buff[i + 1] = (BYTE)(tmp_DATDR >> 16 & 0x000000FF); // Read the response words
            buff[i + 2] = (BYTE)(tmp_DATDR >> 8 & 0x000000FF);  // Read the response words
            buff[i + 3] = (BYTE)(tmp_DATDR & 0x000000FF);       // Read the response words
#endif
            i = i + 4;
        }
    }

    while (i < blockSize) {
#ifdef READ_BY_ASM
        __get_DAT(&MDR_SDIO->DATDR, &buff[i]);
#else
        tmp_DATDR = SDIO->SD_DATDR;
        buff[i] = (BYTE)(tmp_DATDR >> 24 & 0x000000FF);     // Read the response words
        buff[i + 1] = (BYTE)(tmp_DATDR >> 16 & 0x000000FF); // Read the response words
        buff[i + 2] = (BYTE)(tmp_DATDR >> 8 & 0x000000FF);  // Read the response words
        buff[i + 3] = (BYTE)(tmp_DATDR & 0x000000FF);       // Read the response words
#endif
        i = i + 4;
    }
    tmp_DATDR = MDR_SDIO->DATDR;
    tmp_DATDR = MDR_SDIO->DATDR;
    tmp_DATDR = MDR_SDIO->DATDR;

    return true;
}

//  ============= SDIO Write Block ====================

bool SDIO_LL_WriteBlock(const uint8_t* buff)
{
    uint32_t i;

    SDIO_CR_SetMode(SDIO_CR__WriteBlock);

    MDR_SDIO->DAT0CRC = 0x00000000;
    MDR_SDIO->DAT1CRC = 0x00000000;
    MDR_SDIO->DAT2CRC = 0x00000000;
    MDR_SDIO->DAT3CRC = 0x00000000;

    i = 0;
    while (i < 512) {
        MDR_SDIO->DATDR =
            (uint32_t)(buff[i] | buff[i + 1] << 8 | buff[i + 2] << 16 | buff[i + 3] << 24);
        i = i + 4;
    }
    MDR_SDIO->DAT_TRANSFER = 512 * 8 + (16 + 1); // Set length
    MDR_SDIO->CR |= SDIO_CR_WORK1;               // Initiate data transaction

    // Wait for end of the resp transaction
    DelayInt_ms = _SDIO_Timeouts.SD_WriteBlock_Timout_ms;
    while ((MDR_SDIO->CR & SDIO_CR_WORK1) != 0) {
        if (!DelayInt_ms) {
            MDR_SDIO->CR &= ~(SDIO_CR_WORK1 | SDIO_CR_SBITDAT);
            return 0;
        }
        if ((i == 512) && !(MDR_SDIO->SR & SDIO_SR_FIFODAT_FULL)) {
            MDR_SDIO->DATDR = 0x00000000;
            i               = i + 2;
        }
    }

    MDR_SDIO->CR &= ~SDIO_CR_DIRDATA; // DATA RX

    DelayInt_ms = _SDIO_Timeouts.SD_WriteBlock_Timout_ms;
    SDIO_LL_SendClockCycles(8);
    while ((BRD_SDIO_PORT_D0->RXTX & BRD_SDIO_PIN_D0) == 0) {
        SDIO_LL_SendClockCycles(8);
        if (!DelayInt_ms)
            return false;
    }

    return true;
}
